/* 2006-11-18 programmeerimisvistlus                                       */
/* lesanne: SUDOKU                                                         */
/* Autor: Heno Ivanov                                                       */
/* Testitud: GCC 3.2.3 / MinGW32                                            */

/* NB! Loeb standardsisendist, kirjutab standardvljundisse                 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#define XBLOCK 5
#define YBLOCK 3
#define BASESYM 'A'

#define DSIZE XBLOCK*YBLOCK

int numsol = 0;
int rcalls = 0;

struct doku {
    int value;
    int posvalues;
    int vals[DSIZE+1];
};

int cv(struct doku sudoku[DSIZE][DSIZE],int val,int i1,int j1,int i2,int j2) {
    int i,j;
    int c = 0;
    for (i=i1;i<=i2;i++)
        for (j=j1;j<=j2;j++)
            if ((sudoku[i][j].posvalues == 1) && (sudoku[i][j].value == val))
                c++;
    return c;
}

int checkarea(struct doku sudoku[DSIZE][DSIZE],int val,int i,int j) {
    if ((cv(sudoku,val,(i / YBLOCK )*YBLOCK,(j / XBLOCK)*XBLOCK,(i / YBLOCK)*YBLOCK+YBLOCK - 1 ,(j / XBLOCK)*XBLOCK + XBLOCK - 1) == 1) &&
        (cv(sudoku,val,0,j,DSIZE - 1,j) == 1) &&
        (cv(sudoku,val,i,0,i,DSIZE - 1 ) ==  1 )) {
        return 1;
    } else {
        return 0;
    }
}

int excludearea(struct doku sudoku[DSIZE][DSIZE],int val,int i1,int j1,int i2,int j2,int *err) {
    int count = 0;
    int i,j,k;
    for (i=i1;i<=i2;i++)
        for (j=j1;j<=j2;j++) {
            if (sudoku[i][j].posvalues > 1) {
                if (sudoku[i][j].vals[val] == 1) {
                    sudoku[i][j].vals[val] = 0;
                    sudoku[i][j].posvalues--;
                }
                if (sudoku[i][j].posvalues == 1) {
                    for (k=1;k<=DSIZE;k++) {
                        if (sudoku[i][j].vals[k] == 1)
                            sudoku[i][j].value = k;
                    }
                    count++;
                    if (checkarea(sudoku,sudoku[i][j].value,i,j) != 1)
                        *err = 1;
                }
            }
        }
    return count;
}

int exclude(struct doku sudoku[DSIZE][DSIZE],int *err) { // eemaldab reeglite vastuolus olevaid kaiguvariante
    int count = 0;
    int i,j;
    for (i=0;i<DSIZE;i++)
        for (j=0;j<DSIZE;j++) {
            if (sudoku[i][j].posvalues == 1) {
                count+=excludearea(sudoku,sudoku[i][j].value,i,0,i,DSIZE-1,err);
                count+=excludearea(sudoku,sudoku[i][j].value,0,j,DSIZE-1,j,err);
                count+=excludearea(sudoku,sudoku[i][j].value,i - (i % YBLOCK),j - (j % XBLOCK) ,i - (i % YBLOCK) + YBLOCK - 1,j - (j % XBLOCK) + XBLOCK - 1,err);
            }
        }
    return count;
}

int searcharea(struct doku sudoku[DSIZE][DSIZE],int *err,int i1,int j1,int i2,int j2) {
    int count = 0;
    int i,j,k;
    int vals[DSIZE+1];
    int poss[DSIZE+1];
    for (k=1;k<=DSIZE;k++) {
        vals[k] = 0;
        poss[k] = 0;
    }
    for (i=i1;i<=i2;i++)
        for (j=j1;j<=j2;j++)
            if (sudoku[i][j].posvalues == 1) {
                vals[sudoku[i][j].value]++;
                if (vals[sudoku[i][j].value] > 1) {
                    *err = 1;
                    return count;
                }
            }
    for (k=1;k<=DSIZE;k++)
        if (vals[k] == 0) { // k on antud piirkonnas maaramata
            for (i=i1;i<=i2;i++)
                for (j=j1;j<=j2;j++)
                    if ((sudoku[i][j].posvalues > 1) && (sudoku[i][j].vals[k] == 1)) { // leia koht kuhu sobiks
                        poss[k]++;
                    }
        }
    for (k=1;k<=DSIZE;k++) // k on fikseerimata, kuid ei leidu ka yhtegi positsiooni kuhu saaks panna
        if ((vals[k] == 0) && (poss[k] == 0)) {
            *err = 1;
            return count;
        }
    for (k=1;k<=DSIZE;k++)
        if ((vals[k] == 0) && (poss[k] == 1)) // k on fikseerimata, kuid leidub ainult yks positsioon
            for (i=i1;i<=i2;i++)
                for (j=j1;j<=j2;j++)
                    if ((sudoku[i][j].posvalues > 1) && (sudoku[i][j].vals[k] == 1)) {
                        sudoku[i][j].value = k;
                        sudoku[i][j].posvalues = 1;
                        count++;
                    }
    return count;
}

int search(struct doku sudoku[DSIZE][DSIZE],int *err) {
    int count = 0;
    int i,j;
    for (i=0;i<=(XBLOCK - 1);i++)    
        for (j=0;j<=(YBLOCK - 1);j++) {
            count += searcharea(sudoku,err,i*YBLOCK,j*XBLOCK,i*YBLOCK+YBLOCK - 1 ,j*XBLOCK + XBLOCK - 1 );
            if (*err == 1)
                return count;
        }
    for (i=0;i<=(DSIZE-1);i++) {
        count += searcharea(sudoku,err,0,i,DSIZE - 1,i);
        count += searcharea(sudoku,err,i,0,i,DSIZE - 1);
        if (*err == 1)
            return count;
    }
    return count;
}

void pdoku(struct doku sudoku[DSIZE][DSIZE]) {
    int i,j;
    for (i=0;i<DSIZE;i++) {
        for (j=0;j<DSIZE;j++)
            printf("%c",(sudoku[i][j].posvalues == 1) ? (sudoku[i][j].value + BASESYM - 1) : '.');
        printf("\n");
    }
    printf("\n");
}

void rec(struct doku sudoku[DSIZE][DSIZE],int level) {
    struct doku lsudoku[DSIZE][DSIZE];
    int err;
    int i,j;
    int ii,jj;
    err = 0;

    rcalls++;

    for (i=0;i<DSIZE;i++) for (j=0;j<DSIZE;j++) lsudoku[i][j] = sudoku[i][j];
    while (1) {
        int r,s;
        r = exclude(lsudoku,&err);
        s = search(lsudoku,&err);
        if (err == 1)
            return;
        if ((r == 0) && (s==0))
            break;
    }

    for (i=0;i<DSIZE;i++)
        for (j=0;j<DSIZE;j++)
            if (lsudoku[i][j].posvalues >= 2) {
                int k;
                for (k=1;k<=DSIZE;k++)
                    if (lsudoku[i][j].vals[k] == 1) {
                        lsudoku[i][j].posvalues = 1;
                        lsudoku[i][j].value = k;
                        if (checkarea(lsudoku,k,i,j) == 1)
                            rec(lsudoku,level+1);
                        for (ii=0;ii<DSIZE;ii++) for (jj=0;jj<DSIZE;jj++) lsudoku[ii][jj] = sudoku[ii][jj];
                    }
                return;
            }

    numsol++;
    pdoku(lsudoku);
    return;
}

int main() {
    char linebuf[100];
    int i,j,k;
    struct doku sudoku[DSIZE][DSIZE];

    for (i=0;i<DSIZE;i++) {
        fgets(linebuf,20,stdin);
        assert(strlen(linebuf) == DSIZE + 1);
        for (j=0;j<DSIZE;j++) {
            if (linebuf[j] == '.') {
                sudoku[i][j].posvalues = DSIZE;
                for (k=1;k<=DSIZE;k++)
                    sudoku[i][j].vals[k] = 1;
                continue;
            }
            int v = linebuf[j] - BASESYM + 1;
            assert(v>=1 && v<=DSIZE);
            sudoku[i][j].posvalues = 1;
            sudoku[i][j].value = v;
        }
    }
    pdoku(sudoku);

    rec(sudoku,1);
    printf("number of solutions: %d\n",numsol);
    printf("number of nodes: %d\n",rcalls);
    return 0;
}
